/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "JobProperties.hpp"

#include <iostream>
#include <sstream>
#include <map>
#include <cstdlib>

#define PRINT_VAR(x) cout << #x " = " << (x ? x : "NULL") << std::endl

static const bool vfDebug = false;

JobProperties::
JobProperties (PSZCRO pszJobProperties)
{
   setJobProperties (pszJobProperties);
}

JobProperties::
JobProperties (Device *pDevice)
{
   if (pDevice)
   {
      std::string *pstrJobProps = pDevice->queryCurrentJobProperties ();

      if (pstrJobProps)
      {
         PSZCRO pszJobProperties = pstrJobProps->c_str ();

         setJobProperties (pszJobProperties);

         delete pstrJobProps;
      }
   }
}

JobProperties::
~JobProperties ()
{
}

JobProperties& JobProperties::
operator= (const JobProperties& const_jp)
{
   JobProperties& jp      = const_cast<JobProperties&>(const_jp);
   DJPMap&        newDJPs = jp.getDJPMap ();

   mapDJP_d.clear ();

   for ( DJPMap::iterator next = newDJPs.begin ();
         next != newDJPs.end () ;
         next++ )
   {
      mapDJP_d[next->first] = next->second;
   }

   return *this;
}

JobPropertyEnumerator * JobProperties::
getEnumeration (PSZCRO pszKey)
{
   JobPropertyEnumerator *pRet             = 0;
   PSZCRO                 pszJobProperties = getJobProperties ();

   pRet = new JobPropertyEnumerator (pszJobProperties, pszKey);

   if (pszJobProperties)
   {
      free ((void *)pszJobProperties);
   }

   return pRet;
}

PSZCRO JobProperties::
getJobProperties ()
{
   std::ostringstream oss;
   bool               fFirstItem = true;

   for ( DJPMap::iterator next = mapDJP_d.begin ();
         next != mapDJP_d.end () ;
         next++ )
   {
      if (vfDebug) std::cerr << "Adding: " << next->first << "=\"" << next->second << "\"" << std::endl; // @TBD

      if (fFirstItem)
      {
         fFirstItem = false;
      }
      else
      {
         oss << " ";
      }

      oss << next->first
          << "=";

      if (  std::string::npos != next->second.find (' ')
         && '\"' != next->second[0]
         )
      {
         oss << '\"'
             << next->second
             << '\"';
      }
      else
      {
         oss << next->second;
      }
   }

   std::string stringReturn = oss.str ();
   PSZCRO      pszFrom      = stringReturn.c_str ();

   if (  pszFrom
      && *pszFrom
      )
   {
      char *pszTo = 0;

      if (vfDebug) std::cerr << "Returning \"" << pszFrom << "\"" << std::endl; // @TBD

      pszTo = (char *)malloc (strlen (pszFrom) + 1);
      if (pszTo)
      {
         strcpy (pszTo, pszFrom);
      }

      return pszTo;
   }
   else
   {
      return 0;
   }
}

void JobProperties::
setJobProperty (PSZCRO pszKeyValue)
{
   if (  pszKeyValue
      && *pszKeyValue
      )
   {
      std::string            stringKeyValue (pszKeyValue);
      std::string::size_type idx                          = std::string::npos;

      idx = stringKeyValue.find ('=');

      if (std::string::npos != idx)
      {
         setJobProperty (stringKeyValue.substr (0, idx).c_str (),
                         stringKeyValue.substr (idx + 1).c_str ());
      }
   }
}

void JobProperties::
setJobProperty (PSZCRO pszKey,
                PSZCRO pszValue)
{
   if (  pszKey
      && *pszKey
      && pszValue
      && *pszValue
      )
   {
      mapDJP_d[std::string (pszKey)] = std::string (pszValue);
   }
}

void JobProperties::
setJobProperties (PSZCRO pszJobProperties)
{
   if (  pszJobProperties
      && *pszJobProperties
      )
   {
      JobPropertyEnumerator *pEnum = 0;

      pEnum = new JobPropertyEnumerator (pszJobProperties);

      while (pEnum->hasMoreElements ())
      {
         PSZCRO pszKey   = pEnum->getCurrentKey ();
         PSZCRO pszValue = pEnum->getCurrentValue ();

         if (vfDebug) std::cerr << "Setting: " << pszKey << "=\"" << pszValue << "\"" << std::endl; // @TBD

         mapDJP_d[std::string (pszKey)] = std::string (pszValue);

         pEnum->nextElement ();
      }

      delete pEnum;
   }
}

void JobProperties::
applyAllDebugOutput ()
{
   for ( DJPMap::iterator next = mapDJP_d.begin ();
         next != mapDJP_d.end () ;
         next++ )
   {
      if (vfDebug) std::cerr << "DebugOutput: " << next->first << "=\"" << next->second << "\"" << std::endl; // @TBD

      if (next->first.compare ("debugoutput"))
      {
         DebugOutput::setDebugOutput (next->second.c_str ());
      }
   }
}

JobProperties::DJPMap& JobProperties::
getDJPMap ()
{
   return mapDJP_d;
}

JobPropertyEnumerator::
JobPropertyEnumerator (PSZRO  pszJobProperties,
                       PSZCRO pszKey)
{
   pszJobProperties_d   = 0;
   pszKey_d             = 0;
   pszCurrentProperty_d = 0;
   pszCurrentKey_d      = 0;
   pszCurrentValue_d    = 0;

   if (  pszJobProperties
      && *pszJobProperties
      )
   {
      int cbJobProperties = strlen (pszJobProperties) + 1;

      // Skip double quotes around everything
      while (  0 < cbJobProperties
            && '"' == *pszJobProperties
            && '"' == *(pszJobProperties + cbJobProperties - 2)
            )
      {
         pszJobProperties++;
         cbJobProperties -= 2;
      }

      if (0 < cbJobProperties)
      {
         pszJobProperties_d = (char *)malloc (cbJobProperties);

         if (pszJobProperties_d)
         {
            strncpy (pszJobProperties_d, pszJobProperties, cbJobProperties);
            pszJobProperties_d[cbJobProperties - 1] = '\0';

            pszCurrentProperty_d = pszJobProperties_d;
         }
      }
   }
   if (  pszKey
      && *pszKey
      )
   {
      pszKey_d = (char *)malloc (strlen (pszKey) + 1);

      if (pszKey_d)
      {
         strcpy (pszKey_d, pszKey);
      }
   }

///PRINT_VAR (pszJobProperties_d);
///PRINT_VAR (pszKey_d);
///PRINT_VAR (pszCurrentProperty_d);

   nextElement ();
}

JobPropertyEnumerator::
~JobPropertyEnumerator ()
{
   if (pszJobProperties_d)
   {
      free (pszJobProperties_d);
      pszJobProperties_d = 0;
   }
   if (pszKey_d)
   {
      free (pszKey_d);
      pszKey_d = 0;
   }
}

bool JobPropertyEnumerator::
hasMoreElements ()
{
   if (  pszCurrentKey_d
      && pszCurrentValue_d
      )
   {
      return true;
   }
   else
   {
      return false;
   }
}

void JobPropertyEnumerator::
nextElement ()
{
   do
   {
      readNextKey ();

//////PRINT_VAR (pszKey_d);
//////PRINT_VAR (pszCurrentProperty_d);
//////PRINT_VAR (pszCurrentKey_d);
//////PRINT_VAR (pszCurrentValue_d);
//////cout << "hasMoreElements () = " << hasMoreElements () << std::endl;
//////cout << "---------------------" << std::endl;
   } while (  hasMoreElements ()
           && (  pszKey_d
              && 0 != strcmp (pszCurrentKey_d, pszKey_d)
              )
           );
}

PSZCRO JobPropertyEnumerator::
getCurrentKey ()
{
   return pszCurrentKey_d;
}

PSZCRO JobPropertyEnumerator::
getCurrentValue ()
{
   return pszCurrentValue_d;
}

void JobPropertyEnumerator::
readNextKey ()
{
   pszCurrentKey_d   = 0;
   pszCurrentValue_d = 0;

   while (  pszCurrentProperty_d
         && *pszCurrentProperty_d
         && !pszCurrentKey_d
         && !pszCurrentValue_d
         )
   {
      char *pszSep    = pszCurrentProperty_d;
      char *pszEquals = 0;

      // Skip white space
      while (  ' '  == *pszSep
            || '\t' == *pszSep
            || '\r' == *pszSep
            || '\n' == *pszSep
            )
         *pszSep++ = '\0';

      pszEquals = strpbrk (pszSep, "=");

      if (pszEquals)
      {
         *pszEquals = '\0';
         pszCurrentKey_d   = pszSep;
         pszCurrentValue_d = pszEquals + 1;

         if ('\"' == *pszCurrentValue_d)
         {
            pszCurrentValue_d++;

            pszSep = pszCurrentValue_d;
            while (  *pszSep
                  && '\"' != *pszSep
                  )
            {
               pszSep++;
            }

            if ('\"' == *pszSep)
            {
               *pszSep++ = '\0';
            }
            else
            {
               // Error
               pszSep            = 0;
               pszCurrentKey_d   = 0;
               pszCurrentValue_d = 0;
            }
         }
         else
         {
            pszSep = strpbrk (pszCurrentValue_d, " \n\r\t");

            if (pszSep)
            {
               *pszSep++ = '\0';
            }
         }
      }
      else
      {
         if (vfDebug) std::cerr << "Skipping: " << pszSep << std::endl; // @TBD

         pszSep = strpbrk (pszSep, " \n\r\t");

         if (pszSep)
         {
            *pszSep++ = '\0';
         }
      }

      if (  pszSep
         && *pszSep
         )
         pszCurrentProperty_d = pszSep;
      else
         pszCurrentProperty_d = 0;
   }
}
