/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "Device.hpp"
#include "Omni.hpp"
#include "DebugOutput.hpp"

#include <cstdio>
#include <cstdlib>
#include <list>
#include <set>
#include <fstream>

#include <glob.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <sys/stat.h>

PSZCRO vapszLibraryPaths[] = {
   "/opt/Omni/lib/",
   "/usr/lib/Omni/",
   "",                 // give a chance for $LD_LIBRARY_PATH to work
   NULL
};
PSZCRO vapszDataPaths[] = {
   "",                 // allow for a fully qualified filename to work
   "/opt/Omni/share/",
   "/usr/share/Omni/",
   NULL
};

void Omni::
initialize ()
{
#if 0
   char **apszEnviron = environ;

   while (*apszEnviron)
   {
      if (0 != strstr (*apszEnviron, "OMNI"))
      {
         std::cerr << *apszEnviron << std::endl;
      }

      apszEnviron++;
   }
#endif

   DebugOutput::applyDebugEnvironment ();
}

void Omni::
terminate ()
{
}

int Omni::
my_system (PSZCRO pszCommand)
{
   int pid, status;

   if (pszCommand == 0)
      return 1;

   pid = fork();
   if (pid == -1)
      return -1;

   if (pid > 0)
   {
      // parent
      do
      {
         if (waitpid (pid, &status, 0) == -1)
         {
            if (errno != EINTR)
               return -1;
         }
         else
         {
            return status;
         }

      } while (1);
   }
   else
   {
      // child
      char *argv[4];

      argv[0] = "sh";
      argv[1] = "-c";
      argv[2] = (char *)pszCommand;
      argv[3] = 0;

      execvp ("/bin/sh", argv);

      exit (1);
   }
}

void Omni::
setOutputStream (Device *pDevice,
                 FILE   *pfpNew)
{
   if (  pDevice
      && pfpNew
      )
      pDevice->setOutputStream (pfpNew);
}

void Omni::
setErrorStream (Device *pDevice,
                FILE   *pfpNew)
{
   if (!pfpNew)
   {
      return;
   }

   DebugOutput::setErrorStream (pfpNew);

   if (pDevice)
   {
      pDevice->setErrorStream (pfpNew);
   }
}

bool Omni::
libraryValid (GModule *hDevice,
              PSZCRO   pszLibraryName,
              PSZCRO   pszVersion,
              bool     fVerbose)
{
   bool     rc  = false;
   gpointer ptr = NULL;

#ifndef RETAIL
   if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": libraryValid (" << pszLibraryName << ", " << std::hex << hDevice << std::dec << ", " << pszVersion << ")" << std::endl;
#endif

   // The library must have the following symbols in it

   if (  g_module_symbol (hDevice, "getVersion", &ptr)
      && g_module_symbol (hDevice, "getDeviceEnumeration", &ptr)
      && g_module_symbol (hDevice, "newDeviceW_Advanced", &ptr)
      && g_module_symbol (hDevice, "newDeviceW_JopProp_Advanced", &ptr)
      && g_module_symbol (hDevice, "deleteDevice", &ptr)
      )
   {
      PFNGETVERSION pfnGetVersion = 0;

      // Should work.  Already tested above.
      g_module_symbol (hDevice, "getVersion", (gpointer *)&pfnGetVersion);

      PSZCRO pszLibraryVersion = pfnGetVersion ();

      if (  pszLibraryVersion
         && 0 == strcmp (VERSION, pszLibraryVersion)
         )
      {
         rc = true;
      }
#ifndef RETAIL
      else if (fVerbose)
      {
         if (!pszLibraryVersion)
            DebugOutput::getErrorStream () << "for " << pszLibraryName << " getVersion returns null!" << std::endl;
         if (0 != strcmp (VERSION, pszLibraryVersion))
            DebugOutput::getErrorStream () << "for " << pszLibraryName << " " << VERSION << " != " << pszLibraryVersion << std::endl;
      }
#endif
   }
#ifndef RETAIL
   else if (  fVerbose
           && DebugOutput::shouldOutputOmni ()
           )
   {
      g_module_symbol (hDevice, "getDeviceEnumeration", &ptr);
      if (!ptr)
         DebugOutput::getErrorStream () << "for " << pszLibraryName << " getDeviceEnumeration is missing." << std::endl;
      g_module_symbol (hDevice, "newDeviceW_Advanced", &ptr);
      if (!ptr)
         DebugOutput::getErrorStream () << "for " << pszLibraryName << " newDeviceW_Advanced is missing." << std::endl;
      g_module_symbol (hDevice, "newDeviceW_JopProp_Advanced", &ptr);
      if (!ptr)
         DebugOutput::getErrorStream () << "for " << pszLibraryName << " newDeviceW_JopProp_Advanced is missing." << std::endl;
      g_module_symbol (hDevice, "deleteDevice", &ptr);
      if (!ptr)
         DebugOutput::getErrorStream () << "for " << pszLibraryName << " deleteDevice is missing." << std::endl;
      g_module_symbol (hDevice, "getVersion", &ptr);
      if (!ptr)
         DebugOutput::getErrorStream () << "for " << pszLibraryName << " getVersion is missing." << std::endl;
   }

   if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": returning " << rc << std::endl;
#endif

   return rc;
}

static bool
createLibName (char   *pszLibName,
               int     cbLibName,
               PSZCRO  pszDriver,
               PSZRO   pszDevice)
{
   int cbDriver = strlen (pszDriver);

   // Is there enough space for everything?
   if ( cbLibName < (int)( 3                   // "lib"
                         + cbDriver            // driver
                         + 1                   // "_"
                         + strlen (pszDevice)  // device
                         + 3                   // ".so"
                         )
      )
      return false;

   strcpy (pszLibName, "lib");
   strcat (pszLibName, pszDriver);
   strcat (pszLibName, "_");

   char *pszStart = pszLibName + strlen (pszLibName);

   // For example, the driver "Epson" with the device "Epson LQ-2550" is
   // called libEpson_LQ_2550.so.  The "Epson" in the device is redundant.
   if (  0 == strncasecmp (pszDevice,
                           pszDriver,
                           cbDriver)
      && ' ' == pszDevice[cbDriver]
      )
   {
      pszDevice += cbDriver + 1;
   }

   strcat (pszLibName, pszDevice);

   while (*pszStart)
   {
      switch (*pszStart)
      {
      case '-':
      case '+':
      case '(':
      case ')':
         *pszStart = '_';
         break;
      }

      pszStart++;
   }

   strcat (pszLibName, ".so");

   return true;
}

bool Omni::
openLibrary (PSZCRO    pszLibName,
             GModule **phLibrary)
{
#ifndef RETAIL
   if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": g_module_open (" << pszLibName << ")" << std::endl;
#endif

   if (!phLibrary)
   {
      return false;
   }

   *phLibrary = 0;

   if (!g_module_supported ())
   {
      DebugOutput::getErrorStream () << "This program needs glib's module routines!" << std::endl;

      return false;
   }

   char *pszDeviceLib = 0;

   for (int i = 0; vapszLibraryPaths[i] && !*phLibrary; i++)
   {
      pszDeviceLib = (char *)malloc ( strlen (pszLibName)
                                    + strlen (vapszLibraryPaths[i])
                                    + 1);
      if (pszDeviceLib)
      {
         sprintf (pszDeviceLib, "%s%s", vapszLibraryPaths[i], pszLibName);

         *phLibrary = g_module_open (pszDeviceLib, (GModuleFlags)0);
      }
      else
      {
         return false;
      }

      free (pszDeviceLib);
   }

   if (!*phLibrary)
   {
      DebugOutput::getErrorStream () << std::endl << "<<<<<<<<<<<<<<<<<<<<<< ERROR >>>>>>>>>>>>>>>>>>>>>>>" << std::endl;
      DebugOutput::getErrorStream () << std::endl << std::endl;
      DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": for \"" << pszLibName << "\", g_module_error returns \"" << g_module_error () << "\"" << std::endl;
      DebugOutput::getErrorStream () << std::endl;
      DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": Omni device library not found in the following paths:" << std::endl;
      for (int i = 0; vapszLibraryPaths[i]; i++)
      {
         DebugOutput::getErrorStream () << "\t" << vapszLibraryPaths[i] << "." << std::endl;
      }
      DebugOutput::getErrorStream () << "\t$LD_LIBRARY_PATH (" << std::endl;
      char *pszPath = getenv ("LD_LIBRARY_PATH");
      if (pszPath)
      {
         DebugOutput::getErrorStream () << pszPath;
      }
      else
      {
         DebugOutput::getErrorStream () << "NULL";
      }
      DebugOutput::getErrorStream () << ")" << std::endl;

      return 0;
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": phLibrary = " << std::hex << (int)phLibrary << std::dec << std::endl;
#endif

   if (!*phLibrary)
      return false;

   return true;
}

bool Omni::
openDeviceLibrary (PSZCRO    pszLibName,
                   GModule **phDevice)
{
   if (!openLibrary (pszLibName, phDevice))
      return false;

   return libraryValid (*phDevice, pszLibName, VERSION);
}

Device * Omni::
createDevice (PSZCRO    pszDriver,
              PSZCRO    pszDevice,
              PSZCRO    pszJobProperties,
              bool      fAdvanced,
              GModule **phDevice)
{
   char               achLibDeviceName[256]; // @TBD - need max path size
   PFNNEWDEVICEWARGS  pfnNewDeviceWArgs     = 0;
   Device            *pDevice               = 0;

   *phDevice = 0;

   if (!g_module_supported ())
   {
      DebugOutput::getErrorStream () << "This program needs glib's module routines!" << std::endl;
      return 0;
   }

   if (createLibName (achLibDeviceName,
                      sizeof (achLibDeviceName),
                      pszDriver,
                      pszDevice))
   {
      if (openDeviceLibrary (achLibDeviceName, phDevice))
      {
         // Should work.  openDeviceLibrary tests for existance of all required entry points
         g_module_symbol (*phDevice, "newDeviceW_JopProp_Advanced", (gpointer *)&pfnNewDeviceWArgs);

         pDevice = pfnNewDeviceWArgs (pszJobProperties,
                                      fAdvanced);

         if (pDevice)
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputOmni ())
#endif
            {
               DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": strcasecmp (" << pszDriver << ", " << pDevice->getDriverName () << ") = " << std::dec << strcasecmp (pszDriver, pDevice->getDriverName ()) << std::endl;
               DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": strcasecmp (" << pszDevice << ", " << pDevice->getDeviceName () << ") = " << strcasecmp (pszDevice, pDevice->getDeviceName ()) << std::endl;
            }

            if (  0 != strcasecmp (pszDriver, pDevice->getDriverName ())
               && 0 != strcasecmp (pszDevice, pDevice->getDeviceName ())
               )
            {
               // The driver and device name are not the same!
               delete pDevice;
               pDevice = 0;

#ifndef RETAIL
               int rc =
#endif
                        g_module_close (*phDevice);

#ifndef RETAIL
               if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": createDevice @ " << std::dec << __LINE__ << " : g_module_close (" << std::hex << *phDevice << ") = " << rc << std::endl;
#endif

               *phDevice = 0;
            }
         }
      }
   }

   if (!pDevice)
   {
      // Do it the hard way
      Enumeration *pEnum  = Omni::listDevices ();
      bool         fFound = false;

      while (  !fFound
            && pEnum->hasMoreElements ()
            )
      {
         OmniDevice *pOD = (OmniDevice *)pEnum->nextElement ();

         if (pOD)
         {
            PSZCRO pszLibName = pOD->getLibraryName ();

#ifndef RETAIL
            if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": createDevice g_module_open (" << pszLibName << ")" << std::endl;
#endif

            *phDevice = g_module_open (pszLibName, (GModuleFlags)0);

            if (  *phDevice
               && libraryValid (*phDevice, pszLibName, VERSION)
               )
            {
               // Should work.  libraryValid tests for existance of all required entry points
               g_module_symbol (*phDevice, "newDeviceW_JopProp_Advanced", (gpointer *)&pfnNewDeviceWArgs);

               pDevice = pfnNewDeviceWArgs (pszJobProperties, fAdvanced);

#ifndef RETAIL
               if (DebugOutput::shouldOutputOmni ())
               {
                  DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": strcasecmp (" << pszDriver << ", " << pDevice->getDriverName () << ") = " << std::dec << strcasecmp (pszDriver, pDevice->getDriverName ()) << std::endl;
                  DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": strcasecmp (" << pszDevice << ", " << pDevice->getDeviceName () << ") = " << strcasecmp (pszDevice, pDevice->getDeviceName ()) << std::endl;
               }
#endif

               if (  0 == strcasecmp (pszDriver, pDevice->getDriverName ())
                  && 0 == strcasecmp (pszDevice, pDevice->getDeviceName ())
                  )
                  fFound = true;
            }

            delete pOD;
         }

         if (!fFound)
         {
            delete pDevice;
            pDevice = 0;

#ifndef RETAIL
            int rc =
#endif
                     g_module_close (*phDevice);

#ifndef RETAIL
            if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": createDevice @ " << std::dec <<  __LINE__ << " : g_module_close (" << std::hex << *phDevice << ") = " << rc << std::endl;
#endif

            *phDevice = 0;
         }
      }

      delete pEnum;
   }

   return pDevice;
}

Device * Omni::
createDevice (PDL      *pPDL,
              PSZCRO    pszJobProperties,
              bool      fAdvanced,
              GModule **phDevice)
{
   // @TBD
   return 0;
}

DeviceInfo * Omni::
findDeviceInfoFromShortName (PSZCRO pszShortName)
{
   typedef std::list <DeviceInfo *> DeviceInfoList;

   Enumeration    *pEnum               = Omni::listDevices ();
   DeviceInfoList  listMatchingDevices;
   bool            fFoundMatch;

   std::cout << "Info: Trying to figure out which device library to use..." << std::endl;

   while (pEnum->hasMoreElements ())
   {
      OmniDevice *pOD = (OmniDevice *)pEnum->nextElement ();

      fFoundMatch = false;

      if (pOD)
      {
         PSZCRO             pszLibName        = pOD->getLibraryName ();
         PSZCRO             pszJobProperties  = pOD->getJobProperties ();
         Device            *pDevice           = 0;
         PFNNEWDEVICEWARGS  pfnNewDeviceWArgs = 0;
         GModule           *hmodDevice        = 0;

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "OmniDeviceOptions::" << __FUNCTION__ << ": Enumerated: \"" << pszLibName << "\", \"" << pszJobProperties << "\"" << std::endl;
#endif

         hmodDevice = g_module_open (pszLibName, (GModuleFlags)0);

         if (hmodDevice)
         {
            g_module_symbol (hmodDevice,
                             "newDeviceW_JopProp_Advanced",
                             (gpointer *)&pfnNewDeviceWArgs);

            if (pfnNewDeviceWArgs)
            {
               pDevice = pfnNewDeviceWArgs (pszJobProperties, true);

               if (pDevice)
               {
                  if (0 == strcmp (pszShortName, pDevice->getShortName ()))
                  {
#ifndef RETAIL
                     if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "OmniDeviceOptions::" << __FUNCTION__ << ": Found a match!" << std::endl;
#endif

                     fFoundMatch = true;

                     listMatchingDevices.push_back (new DeviceInfo (pDevice,
                                                                    hmodDevice,
                                                                    pOD));
                  }

                  if (!fFoundMatch)
                     delete pDevice;
               }
            }

            if (!fFoundMatch)
               g_module_close (hmodDevice);
         }

         if (!fFoundMatch)
            delete pOD;
      }
   }

   if (0 == listMatchingDevices.size ())
   {
      std::cerr << "Error: Could not find a device whose short name matches " << pszShortName << std::endl;
   }
   else if (1 == listMatchingDevices.size ())
   {
      return listMatchingDevices.front ();
   }
   else
   {
      std::cerr << "Error: There are multiple devices whose short names match " << pszShortName << std::endl;

      for ( DeviceInfoList::iterator next = listMatchingDevices.begin () ;
            next != listMatchingDevices.end () ;
            next++ )
      {
         DeviceInfo  *pDI        = *next;
         Device      *pDevice    = 0;
         GModule     *hmodDevice = 0;
         OmniDevice  *pOD        = 0;
         PSZRO        pszJobProp = 0;

         if (pDI)
         {
            pDevice    = pDI->getDevice ();
            hmodDevice = pDI->getHmodDevice ();
            pOD        = pDI->getOmniDevice ();

            if (pOD)
            {
               pszJobProp = pOD->getJobProperties ();
            }

            std::cerr << "Error:\t";

            if (pszJobProp)
            {
               std::cerr << "'-sproperties="
                         << pszJobProp
                         << "'";
            }

            std::cerr << " --driver"
                      << pDevice->getLibraryName ()
                      << std::endl;

            delete pDI;
         }
      }
   }

   delete pEnum;

   return 0;
}

PSZCRO Omni::
openXMLFile (PSZCRO pszXMLFile)
{
   if (  !pszXMLFile
      || !*pszXMLFile
      )
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": pszXMLFile = NULL" << std::endl;
#endif

      return 0;
   }

   char *pszTargetXMLFile = 0;

   for (int i = 0; vapszDataPaths[i]; i++)
   {
      pszTargetXMLFile = (char *)malloc ( strlen (pszXMLFile)
                                        + strlen (vapszDataPaths[i])
                                        + 1
                                        );
      if (pszTargetXMLFile)
      {
         struct stat statFile;

         strcpy (pszTargetXMLFile, vapszDataPaths[i]);
         strcat (pszTargetXMLFile, pszXMLFile);

         if (0 == stat (pszTargetXMLFile, &statFile))
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": Mapping \"" << pszXMLFile << "\" to \"" << pszTargetXMLFile << "\"." << std::endl;
#endif

            return pszTargetXMLFile;
         }

         free (pszTargetXMLFile);
         pszTargetXMLFile = 0;
      }
      else
      {
         DebugOutput::getErrorStream () << std::endl << "<<<<<<<<<<<<<<<<<<<<<< ERROR >>>>>>>>>>>>>>>>>>>>>>>" << std::endl;
         DebugOutput::getErrorStream () << std::endl << std::endl;
         DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": Error: Out of memory @ " << __LINE__ << std::endl;

         return 0;
      }
   }

   DebugOutput::getErrorStream () << std::endl << "<<<<<<<<<<<<<<<<<<<<<< ERROR >>>>>>>>>>>>>>>>>>>>>>>" << std::endl;
   DebugOutput::getErrorStream () << std::endl << std::endl;
   DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": Omni XML file (" << pszXMLFile << ") not found in the following paths:" << std::endl;
   for (int i = 0; vapszDataPaths[i]; i++)
   {
      if (*vapszDataPaths[i])
      {
         DebugOutput::getErrorStream () << "\t" << vapszDataPaths[i] << std::endl;
      }
      else
      {
         DebugOutput::getErrorStream () << "\t." << std::endl;
      }
   }
   DebugOutput::getErrorStream () << std::endl;

   return 0;
}

class BuildDeviceEnumerator : public Enumeration
{
public:
   BuildDeviceEnumerator (bool fVerbose)
   {
      // Initialize our variables
      fVerbose_d           = fVerbose;
      pszLD_LIBRARY_PATH_d = getenv ("LD_LIBRARY_PATH");
      fGlobActive_d        = false;
      hLibrary_d           = 0;
      pDeviceEnum_d        = 0;
   }

   virtual
   ~BuildDeviceEnumerator ()
   {
      // Clean up
      if (fGlobActive_d)
      {
         globfree (&globbuf_d);
         fGlobActive_d = false;
      }

      delete pDeviceEnum_d;
      pDeviceEnum_d = 0;
      if (hLibrary_d)
      {
         g_module_close (hLibrary_d);
         hLibrary_d = 0;
      }
   }

   virtual bool
   hasMoreElements ()
   {
      // Prime the pump.  That is, find at least one element.

      // Do we have an active device enumeration?
      if (pDeviceEnum_d)
      {
         if (pDeviceEnum_d->hasMoreElements ())
         {
            return true;
         }

         delete pDeviceEnum_d;
         pDeviceEnum_d = 0;
         if (hLibrary_d)
         {
            g_module_close (hLibrary_d);
            hLibrary_d = 0;
         }
      }

      // Are we in a glob?
      if (fGlobActive_d)
      {
         // Find another one
         findDeviceInGlob ();
      }

      if (  !fGlobActive_d        // Did we not find one?
         && pszLD_LIBRARY_PATH_d  // Still more paths to search
         )
      {
         // Try with another path
         pszLD_LIBRARY_PATH_d = globPath (pszLD_LIBRARY_PATH_d);
      }

      // Did we find one?
      if (fGlobActive_d)
         return true;
      else
         return false;
   }

   virtual void *
   nextElement ()
   {
      if (pDeviceEnum_d)
      {
         return pDeviceEnum_d->nextElement ();
      }
      else if (fGlobActive_d)
      {
         PSZCRO pszLibraryName = globbuf_d.gl_pathv[iCurrentFile_d];

         // See if this device has subdevices
         hLibrary_d = g_module_open (pszLibraryName, (GModuleFlags)0);

         if (hLibrary_d)
         {
            PFNENUMERATEDEVICES pfnEnumerateDevices = 0;

            g_module_symbol (hLibrary_d,
                             "getDeviceEnumeration",
                             (gpointer *)&pfnEnumerateDevices);

            if (pfnEnumerateDevices)
            {
               pDeviceEnum_d = pfnEnumerateDevices (pszLibraryName, true);
            }

            if (pDeviceEnum_d)
            {
               if (pDeviceEnum_d->hasMoreElements ())
               {
                  // Return the first subdevice
                  return pDeviceEnum_d->nextElement ();
               }
            }

            // Clean up
            delete pDeviceEnum_d;
            pDeviceEnum_d = 0;
            g_module_close (hLibrary_d);
            hLibrary_d = 0;
         }

         // Return the one that we found
         return new OmniDevice (pszLibraryName, 0);
      }
      else
      {
         // Error!
         return 0;
      }
   }

private:
   char *
   globPath (char *pszLD_LIBRARY_PATH)
   {
      char  achPath[256]; // @TBD - need max path size
      char *pszSep       = 0;
      bool  fFound       = false;
      int   rc;

      do
      {
         // See if there are two or more paths
         pszSep = strpbrk (pszLD_LIBRARY_PATH, ":");

         if (pszSep)
         {
            int iChars = pszSep - pszLD_LIBRARY_PATH;

            // Just copy the first one
            strncpy (achPath, pszLD_LIBRARY_PATH, iChars);
            achPath[iChars] = '\0';

            pszLD_LIBRARY_PATH = pszSep + 1;
         }
         else
         {
            // Copy the only one
            strcpy (achPath, pszLD_LIBRARY_PATH);

            pszLD_LIBRARY_PATH = 0;
         }

         // Add our search parameter
         strcat (achPath, "/lib*.so");

         if (fVerbose_d)
         {
            DebugOutput::getErrorStream () << "Searching: glob (" << achPath << ") = ";
         }

         // Call glob
         memset (&globbuf_d, 0, sizeof (globbuf_d));
         rc = glob (achPath, 0, NULL, &globbuf_d);

         if (fVerbose_d)
         {
            DebugOutput::getErrorStream () << rc << std::endl;
         }

         if (0 == rc)
         {
            // Call succeded
            iCurrentFile_d = -1;
            fGlobActive_d  = true;

            // See if there is a device in this directory
            fFound = findDeviceInGlob ();
         }

      } while (  !fFound            // Haven't found any yet
              && pszLD_LIBRARY_PATH // More to search
              );

      return pszLD_LIBRARY_PATH;
   }

   bool
   findDeviceInGlob (void)
   {
      bool fFound = false;

      for ( iCurrentFile_d++;
            iCurrentFile_d < (int)globbuf_d.gl_pathc;
            iCurrentFile_d++)
      {
         PSZCRO pszCurrentName = globbuf_d.gl_pathv[iCurrentFile_d];

         if (fVerbose_d)
         {
            DebugOutput::getErrorStream () << "Checking (" << pszCurrentName << ")" << std::endl;
         }

         std::string stringSystem;

/////////stringSystem  = "strace OmniDeviceValid ";
         stringSystem  = "OmniDeviceValid ";
         if (fVerbose_d)
            stringSystem += "-v ";
         stringSystem += "\"";
         stringSystem += pszCurrentName;
         stringSystem += "\" ";
         stringSystem += VERSION;

         if (0 == Omni::my_system (stringSystem.c_str ()))
         {
            fFound = true;
            break;
         }
      }

      if (!fFound)
      {
         globfree (&globbuf_d);
         fGlobActive_d = false;
      }

      return fFound;
   }

private:
   bool         fVerbose_d;
   char        *pszLD_LIBRARY_PATH_d;
   glob_t       globbuf_d;
   int          iCurrentFile_d;
   bool         fGlobActive_d;

   GModule     *hLibrary_d;
   Enumeration *pDeviceEnum_d;
};

class SystemDeviceEnumerator : public Enumeration
{
public:
   SystemDeviceEnumerator ()
   {
      // Initialize our variables

      // Add the Omni builtin paths
      for (int i = 0; vapszLibraryPaths[i]; i++)
      {
         std::string stringPath = vapszLibraryPaths[i];

         if (*vapszLibraryPaths[i])
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::SystemDeviceEnumerator::" << __FUNCTION__ << ": inserting (" << stringPath << ")" << std::endl;
#endif

            setDirectories_d.insert (stringPath);
         }
      }

      // Add /etc/ld.so.conf
      char           achLine[512];             // @TBD
      std::ifstream  ifIn ("/etc/ld.so.conf");

      while (0 < ifIn.getline (achLine, sizeof (achLine)))
      {
         char *pszEnd = achLine + strlen (achLine) - 1;

         if ('/' != *pszEnd)
         {
            strcat (achLine, "/");
         }

         std::string stringLine = achLine;

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::SystemDeviceEnumerator::" << __FUNCTION__ << ": inserting (" << achLine << ")" << std::endl;
#endif

         setDirectories_d.insert (stringLine);
      }

      iterCurrent_d = setDirectories_d.begin ();

      fGlobActive_d = false;
      hLibrary_d    = 0;
      pDeviceEnum_d = 0;
   }

   virtual
   ~SystemDeviceEnumerator ()
   {
      // Clean up
      if (fGlobActive_d)
      {
         globfree (&globbuf_d);
         fGlobActive_d = false;
      }

      delete pDeviceEnum_d;
      pDeviceEnum_d = 0;
      if (hLibrary_d)
      {
         g_module_close (hLibrary_d);
         hLibrary_d = 0;
      }
   }

   virtual bool
   hasMoreElements ()
   {
      // Prime the pump.  That is, find at least one element.

      // Do we have an active device enumeration?
      if (pDeviceEnum_d)
      {
         if (pDeviceEnum_d->hasMoreElements ())
         {
            return true;
         }

         delete pDeviceEnum_d;
         pDeviceEnum_d = 0;
         if (hLibrary_d)
         {
            g_module_close (hLibrary_d);
            hLibrary_d = 0;
         }
      }

      // Are we in a glob?
      if (fGlobActive_d)
      {
         // Find another one
         findDeviceInGlob ();
      }

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::SystemDeviceEnumerator::" << __FUNCTION__ << ": fGlobActive_d = " << fGlobActive_d << ", " << (iterCurrent_d != setDirectories_d.end ()) << std::endl;
#endif

      if (  !fGlobActive_d                           // Did we not find one?
         && iterCurrent_d != setDirectories_d.end () // Still more directories to search
         )
      {
         // Try with another path
         globPath ();
      }

      // Did we find one?
      if (fGlobActive_d)
         return true;
      else
         return false;
   }

   virtual void *
   nextElement ()
   {
      if (pDeviceEnum_d)
      {
         return pDeviceEnum_d->nextElement ();
      }
      else if (fGlobActive_d)
      {
         PSZCRO pszLibraryName = globbuf_d.gl_pathv[iCurrentFile_d];

         // See if this device has subdevices
         hLibrary_d = g_module_open (pszLibraryName, (GModuleFlags)0);

         if (hLibrary_d)
         {
            PFNENUMERATEDEVICES pfnEnumerateDevices = 0;

            g_module_symbol (hLibrary_d,
                             "getDeviceEnumeration",
                             (gpointer *)&pfnEnumerateDevices);

            if (pfnEnumerateDevices)
            {
               pDeviceEnum_d = pfnEnumerateDevices (pszLibraryName, false);
            }

            if (pDeviceEnum_d)
            {
               if (pDeviceEnum_d->hasMoreElements ())
               {
                  // Return the first subdevice
                  return pDeviceEnum_d->nextElement ();
               }
            }

            // Clean up
            delete pDeviceEnum_d;
            pDeviceEnum_d = 0;
            g_module_close (hLibrary_d);
            hLibrary_d = 0;
         }

         // Return the one that we found
         return new OmniDevice (pszLibraryName, 0);
      }
      else
      {
         // Error!
         return 0;
      }
   }

private:
   void
   globPath ()
   {
      bool fFound = false;
      int  rc     = 0;

      while (  !fFound                                  // Haven't found any yet
            && iterCurrent_d != setDirectories_d.end () // Haven't reached the end
            )
      {
         std::string stringGlob = *iterCurrent_d++;

         stringGlob += "lib*.so";

         // Call glob
         memset (&globbuf_d, 0, sizeof (globbuf_d));
         rc = glob (stringGlob.c_str (), 0, NULL, &globbuf_d);

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::SystemDeviceEnumerator::" << __FUNCTION__ << " (" << stringGlob << ") = " << rc << std::endl;
#endif

         if (0 == rc)
         {
            // Call succeded
            iCurrentFile_d = -1;
            fGlobActive_d  = true;

            // See if there is a device in this directory
            fFound = findDeviceInGlob ();
         }
      }
   }

   bool
   findDeviceInGlob (void)
   {
      bool fFound = false;

      for ( iCurrentFile_d++;
            iCurrentFile_d < (int)globbuf_d.gl_pathc;
            iCurrentFile_d++)
      {
         PSZCRO pszCurrentName = globbuf_d.gl_pathv[iCurrentFile_d];

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::SystemDeviceEnumerator::" << __FUNCTION__ << ": findDeviceInGlob: checking (" << pszCurrentName << ")" << std::endl;
#endif

         std::string stringSystem;

         stringSystem  = "OmniDeviceValid ";
/////////stringSystem += "-v ";
         stringSystem += "\"";
         stringSystem += pszCurrentName;
         stringSystem += "\" ";
         stringSystem += VERSION;

         if (0 == Omni::my_system (stringSystem.c_str ()))
         {
            fFound = true;
            break;
         }
      }

      if (!fFound)
      {
         globfree (&globbuf_d);
         fGlobActive_d = false;
      }

      return fFound;
   }

private:
   std::set<std::string>            setDirectories_d;
   std::set<std::string>::iterator  iterCurrent_d;
   glob_t                           globbuf_d;
   int                              iCurrentFile_d;
   bool                             fGlobActive_d;

   GModule                         *hLibrary_d;
   Enumeration                     *pDeviceEnum_d;
};

Enumeration * Omni::
listDevices (bool fBuildOnly)
{
   if (fBuildOnly)
   {
      return new BuildDeviceEnumerator (true);
   }
   else
   {
      return new SystemDeviceEnumerator ();
   }
}

static bool
checkForXMLMasterSignature (PSZCRO pszFileName)
{
   FILE *fpFile = 0;
   bool  fFound = false;

   fpFile = fopen (pszFileName, "r");

   if (fpFile)
   {
//#define USE_DTD 1
#ifdef USE_DTD
                                    //           1         2          3         4
                                    // 0123456789012345678901234 567890123456789012 34
      static PSZCRO  pszTest        = "<!DOCTYPE Device SYSTEM \"../OmniDevice.dtd\">";
#else
                                    //           1
                                    // 01234567890123
      static PSZCRO  pszTest        = "<Device name=\"";
#endif
      static int     cbTest         = strlen (pszTest);
      char           achBuffer[512];
      char          *pszBuffer      = achBuffer;
      size_t         cbBuffer       = sizeof (achBuffer);
      bool           fContinue      = true;

      do
      {
         // Read a block
         if (fread (pszBuffer, cbBuffer, 1, fpFile))
         {
            // Start at the end
            char *pszSearch = achBuffer + sizeof (achBuffer) - cbTest;

            // While within the buffer
            while (pszSearch >= achBuffer)
            {
               // Is is found
               if (  *pszSearch == *pszTest
                  && 0 == strncmp (pszSearch, pszTest, cbTest)
                  )
               {
                  fFound = true;

                  break;
               }
               else
               {
                  // Move backwards based on current character
                  switch (*pszSearch)
                  {
#ifdef USE_DTD
                  case '!': pszSearch -= 1;      break;
                  case 'D': pszSearch -= 2;      break;
                  case 'O': pszSearch -= 3;      break;
                  case 'C': pszSearch -= 4;      break;
                  case 'T': pszSearch -= 5;      break;
                  case 'Y': pszSearch -= 6;      break;
                  case 'P': pszSearch -= 7;      break;
                  case 'E': pszSearch -= 8;      break;
                  case ' ': pszSearch -= 9;      break;
                  case 'e': pszSearch -= 11;     break;
                  case 'v': pszSearch -= 12;     break;
                  case 'i': pszSearch -= 13;     break;
                  case 'c': pszSearch -= 14;     break;
                  case 'S': pszSearch -= 17;     break;
                  case 'M': pszSearch -= 22;     break;
                  case '"': pszSearch -= 24;     break;
                  case '/': pszSearch -= 27;     break;
                  case 'm': pszSearch -= 29;     break;
                  case 'n': pszSearch -= 30;     break;
                  case '.': pszSearch -= 38;     break;
                  case 'd': pszSearch -= 39;     break;
                  case 't': pszSearch -= 40;     break;
                  case '>': pszSearch -= 44;     break;
                  default:  pszSearch -= cbTest; break;
#else
                  case 'D': pszSearch -= 1;      break;
                  case 'e': pszSearch -= 2;      break;
                  case 'v': pszSearch -= 3;      break;
                  case 'i': pszSearch -= 4;      break;
                  case 'c': pszSearch -= 5;      break;
                  case ' ': pszSearch -= 7;      break;
                  case 'n': pszSearch -= 8;      break;
                  case 'a': pszSearch -= 9;      break;
                  case 'm': pszSearch -= 10;     break;
                  case '=': pszSearch -= 12;     break;
                  case '"': pszSearch -= 13;     break;
                  default:  pszSearch -= cbTest; break;
#endif
                  }
               }
            }

            if (!fFound)
            {
               /* In order to handle the case where the search string might
               ** be spanning buffers, move the end of the buffer to the
               ** front and shrink the buffer
               */
               memmove (achBuffer,
                        achBuffer + sizeof (achBuffer) - cbTest,
                        cbTest);

               pszBuffer = achBuffer + cbTest;
               cbBuffer  = sizeof (achBuffer) - cbTest;
            }
         }
         else
         {
            // File read error!
            fContinue = false;
         }

      } while (  fContinue
              && !fFound
              );

      // Clean up
      fclose (fpFile);
   }

   return fFound;
}

class PathXMLEnumerator : public Enumeration
{
public:
   PathXMLEnumerator ()
   {
      // Initialize our variables
      iterCurrent_d = setDirectories_d.begin ();

      fGlobActive_d = false;
   }

   virtual
   ~PathXMLEnumerator ()
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::PathXMLEnumerator::~" << __FUNCTION__ << std::endl;
#endif

      // Clean up
      if (fGlobActive_d)
      {
         globfree (&globbuf_d);
         fGlobActive_d = false;
      }
   }

   virtual bool
   hasMoreElements ()
   {
      // Prime the pump.  That is, find at least one element.

      // Are we in a glob?
      if (fGlobActive_d)
      {
         // Find another one
         findDeviceInGlob ();
      }

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::PathXMLEnumerator::" << __FUNCTION__ << ": fGlobActive_d = " << fGlobActive_d << ", " << (iterCurrent_d != setDirectories_d.end ()) << std::endl;
#endif

      if (  !fGlobActive_d                           // Did we not find one?
         && iterCurrent_d != setDirectories_d.end () // Still more directories to search
         )
      {
         // Try with another path
         globPath ();
      }

      // Did we find one?
      if (fGlobActive_d)
         return true;
      else
         return false;
   }

   virtual void *
   nextElement ()
   {
      if (fGlobActive_d)
         // Return the one that we found
         return globbuf_d.gl_pathv[iCurrentFile_d];
      else
         // Error!
         return 0;
   }

   void
   addPath (std::string stringPath)
   {
      setDirectories_d.insert (stringPath);

      iterCurrent_d = setDirectories_d.begin ();

      fGlobActive_d = false;
   }

private:
   void
   globPath ()
   {
      bool fFound = false;
      int  rc     = 0;

      while (  !fFound                                  // Haven't found any yet
            && iterCurrent_d != setDirectories_d.end () // Haven't reached the end
            )
      {
         std::string stringGlob = *iterCurrent_d++;

         stringGlob += "*.xml";

         // Call glob
         memset (&globbuf_d, 0, sizeof (globbuf_d));
         rc = glob (stringGlob.c_str (), 0, NULL, &globbuf_d);

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::PathXMLEnumerator::" << __FUNCTION__ << " (" << stringGlob << ") = " << rc << std::endl;
#endif

         if (0 == rc)
         {
            // Call succeded
            iCurrentFile_d = -1;
            fGlobActive_d  = true;

            // See if there is a device in this directory
            fFound = findDeviceInGlob ();
         }
      }
   }

   bool
   findDeviceInGlob (void)
   {
      bool fFound = false;

      for ( iCurrentFile_d++;
            iCurrentFile_d < (int)globbuf_d.gl_pathc;
            iCurrentFile_d++)
      {
         PSZCRO pszCurrentName = globbuf_d.gl_pathv[iCurrentFile_d];

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::PathXMLEnumerator::" << __FUNCTION__ << ": findDeviceInGlob: checking (" << pszCurrentName << ")" << std::endl;
#endif

         fFound = checkForXMLMasterSignature (pszCurrentName);

         if (fFound)
         {
            break;
         }
      }

      if (!fFound)
      {
         globfree (&globbuf_d);
         fGlobActive_d = false;
      }

      return fFound;
   }

private:
   std::set<std::string>           setDirectories_d;
   std::set<std::string>::iterator iterCurrent_d;
   glob_t                          globbuf_d;
   int                             iCurrentFile_d;
   bool                            fGlobActive_d;
};

class DeviceListXMLEnumerator : public Enumeration
{
public:
   DeviceListXMLEnumerator ()
   {
      // Initialize our variables
      iterCurrent_d = setDirectories_d.begin ();
      fGlobActive_d = false;
      pifstreamIn_d = 0;
   }

   virtual
   ~DeviceListXMLEnumerator ()
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::DeviceListXMLEnumerator::~" << __FUNCTION__ << std::endl;
#endif

      // Clean up
      delete pifstreamIn_d; pifstreamIn_d = 0;
   }

   virtual bool
   hasMoreElements ()
   {
      // Prime the pump.  That is, find at least one element.

      // Are we in a glob?
      if (fGlobActive_d)
      {
         // Find another one
         findDeviceInGlob ();
      }

#ifndef RETAIL
      if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::DeviceListXMLEnumerator::" << __FUNCTION__ << ": fGlobActive_d = " << fGlobActive_d << ", " << (iterCurrent_d != setDirectories_d.end ()) << std::endl;
#endif

      if (  !fGlobActive_d                           // Did we not find one?
         && iterCurrent_d != setDirectories_d.end () // Still more directories to search
         )
      {
         // Try with another path
         globPath ();
      }

      // Did we find one?
      if (fGlobActive_d)
         return true;
      else
         return false;
   }

   virtual void *
   nextElement ()
   {
      if (fGlobActive_d)
      {
         // Return the one that we found
         return (void *)stringMasterXMLFileName_d.c_str ();
      }
      else
      {
         // Error!
         return 0;
      }
   }

   void
   addPath (std::string stringPath)
   {
      setDirectories_d.insert (stringPath);

      iterCurrent_d = setDirectories_d.begin ();

      fGlobActive_d = false;
   }

private:
   void
   globPath ()
   {
      bool fFound = false;

      while (  !fFound                                  // Haven't found any yet
            && iterCurrent_d != setDirectories_d.end () // Haven't reached the end
            )
      {
         std::string stringFileName = *iterCurrent_d;
         struct stat statFileName;

         stringFileName += "Device List";

         delete pifstreamIn_d; pifstreamIn_d = 0;

         if (-1 != stat (stringFileName.c_str (), &statFileName))
         {
            pifstreamIn_d = new std::ifstream (stringFileName.c_str ());

            // See if there is a device in this directory
            fFound = findDeviceInGlob ();

            if (fFound)
            {
               // Call succeded
               fGlobActive_d = true;
            }
         }

         if (!fFound)
         {
            iterCurrent_d++;
         }
      }
   }

   bool
   findDeviceInGlob (void)
   {
      char  achLine[512];   // @TBD
      bool  fFound       = false;

      if (!pifstreamIn_d)
      {
         return false;
      }

      while (0 < pifstreamIn_d->getline (achLine, sizeof (achLine)))
      {
         if ('#' == achLine[0])
            continue;

         std::string stringFileNameMasterXML;

         stringFileNameMasterXML  = *iterCurrent_d;
         stringFileNameMasterXML += achLine;
         stringFileNameMasterXML += ".xml";

#ifndef RETAIL
         if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::DeviceListXMLEnumerator::" << __FUNCTION__ << ": findDeviceInGlob: checking (" << stringFileNameMasterXML << ")" << std::endl;
#endif

         fFound = checkForXMLMasterSignature (stringFileNameMasterXML.c_str ());

         if (fFound)
         {
            stringMasterXMLFileName_d = stringFileNameMasterXML;
            break;
         }
      }

      if (!fFound)
      {
         fGlobActive_d = false;

         iterCurrent_d++;
      }

      return fFound;
   }

private:
   std::set<std::string>            setDirectories_d;
   std::set<std::string>::iterator  iterCurrent_d;
   bool                             fGlobActive_d;
   std::ifstream                   *pifstreamIn_d;
   std::string                      stringMasterXMLFileName_d;
};

Enumeration * Omni::
listXMLDevices (bool fBuildOnly)
{
   if (fBuildOnly)
   {
      DeviceListXMLEnumerator *pEnum       = 0;
      PSZCRO                   pszXML_PATH = getenv ("OMNI_DEVICELIST_PATH");

      if (!pszXML_PATH)
      {
#ifndef RETAIL
         if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": Error OMNI_DEVICELIST_PATH is not set!" << std::endl;
#endif
      }

      pEnum = new DeviceListXMLEnumerator ();

      if (!pEnum)
      {
#ifndef RETAIL
         if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": Error failed to allocate DeviceListXMLEnumerator!" << std::endl;
#endif
      }

      if (  pEnum
         && pszXML_PATH
         )
      {
         std::string            stringXMLPath = pszXML_PATH;
         std::string::size_type indexCurrent  = 0;
         std::string::size_type indexNext     = 0;

         do
         {
            std::string stringPath;

            // See if there are two or more paths
            indexNext = stringXMLPath.find (":", indexCurrent);

            if (std::string::npos != indexNext)
            {
               stringPath = stringXMLPath.substr (indexCurrent,
                                                  indexNext - indexCurrent);
            }
            else
            {
               stringPath = stringXMLPath.substr (indexCurrent);
            }

            if (0 < stringPath.length ())
            {
               if (stringPath[stringPath.length () - 1] != '/')
               {
                  stringPath += "/";
               }

#ifndef RETAIL
               if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": inserting (" << stringPath << ")" << std::endl;
#endif

               pEnum->addPath (stringPath);
            }

            indexCurrent = indexNext + 1;

         } while (std::string::npos != indexNext); // More to search
      }

      return pEnum;
   }
   else
   {
      PathXMLEnumerator *pEnum = 0;

      pEnum = new PathXMLEnumerator ();

      if (pEnum)
      {
         // Add the Omni builtin paths
         for (int i = 0; vapszDataPaths[i]; i++)
         {
            std::string stringPath = vapszDataPaths[i];

            if (*vapszDataPaths[i])
            {
#ifndef RETAIL
               if (DebugOutput::shouldOutputOmni ()) DebugOutput::getErrorStream () << "Omni::" << __FUNCTION__ << ": inserting (" << stringPath << ")" << std::endl;
#endif

               pEnum->addPath (stringPath);
            }
         }
      }

      return pEnum;
   }

   return 0;
}
