/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "Device.hpp"
#include "Omni.hpp"
#include "DeviceInfo.hpp"
#include "JobProperties.hpp"

#include <cstdio>
#include <cstdlib>
#include <memory>
#include <string>
#include <list>

#include <glib.h>
#include <gmodule.h>

typedef int (*PFNSHOWOPTIONS) (Device *pDevice);

void
printUsage (char *pszProgramName)
{
   std::cerr << "Usage: "
             << pszProgramName
             << " ( --UsePDC | --!UsePDC | --driver printer_library_name | '-sproperties=\"...\"' | --Verbose | --all | printer_short_name )+"
             << std::endl;
}

int
executeDevice (PSZRO             pszFullDeviceName,
               bool              fUsePDC,
               PFNSHOWOPTIONS    pfnShowOptions,
               PSZCRO            pszJobProperties)
{
   Device       *pDevice      = 0;
   GModule      *hmodDevice   = 0;
   int           rc           = 0;

#ifndef RETAIL
   if (DebugOutput::shouldOutputDeviceTester ()) DebugOutput::getErrorStream () << "OmniDeviceOptions::" << __FUNCTION__ << ": pszFullDeviceName = \"" << pszFullDeviceName << "\" pszJobProperties = \"" << pszJobProperties << "\"" << std::endl;
#endif

   if (fUsePDC)
   {
      char *pszSlash = strchr (pszFullDeviceName, '/');

      while (pszSlash)
      {
         pszFullDeviceName = pszSlash + 1;

         pszSlash = strchr (pszFullDeviceName, '/');
      }

      pDevice = new OmniPDCProxy (0,                 // client exe to spawn
                                  pszFullDeviceName, // device name
                                  pszJobProperties,  // job properties
                                  true);             // is renderer
   }
   else
   {
      PFNNEWDEVICE       pfnNewDevice      = 0;
      PFNNEWDEVICEWARGS  pfnNewDeviceWArgs = 0;

      hmodDevice = g_module_open (pszFullDeviceName, (GModuleFlags)0);

#ifndef RETAIL
      if (DebugOutput::shouldOutputDeviceTester ()) DebugOutput::getErrorStream () << "OmniDeviceOptions::" << __FUNCTION__ << ": hmodDevice        = " << std::hex << hmodDevice << std::dec << std::endl;
#endif

      if (!hmodDevice)
      {
         DebugOutput::getErrorStream ()
            << "Error: trying to load \""
            << pszFullDeviceName
            << "\" g_module_error returns "
            << g_module_error ()
            << std::endl;

         return 2;
      }

      g_module_symbol (hmodDevice, "newDeviceW_Advanced",         (gpointer *)&pfnNewDevice);
      g_module_symbol (hmodDevice, "newDeviceW_JopProp_Advanced", (gpointer *)&pfnNewDeviceWArgs);

#ifndef RETAIL
      if (DebugOutput::shouldOutputDeviceTester ()) DebugOutput::getErrorStream () << "OmniDeviceOptions::" << __FUNCTION__ << std::hex << ": pfnNewDevice      = 0x" << (int)pfnNewDevice << std::endl;
      if (DebugOutput::shouldOutputDeviceTester ()) DebugOutput::getErrorStream () << "OmniDeviceOptions::" << __FUNCTION__ << std::hex << ": pfnNewDeviceWArgs = 0x" << (int)pfnNewDeviceWArgs << std::endl;
#endif

      if (  !pfnNewDevice
         || !pfnNewDeviceWArgs
         )
      {
         DebugOutput::getErrorStream () << "Error: trying to load newDeviceW[_JobProp]_Advanced. g_module_error returns " << g_module_error () << std::endl;

         return 3;
      }

      if (  !pszJobProperties
         || !*pszJobProperties
         )
      {
         pDevice = pfnNewDevice (true);
      }
      else
      {
         pDevice = pfnNewDeviceWArgs (pszJobProperties, true);
      }
   }

#ifndef RETAIL
   if (DebugOutput::shouldOutputDeviceTester ()) DebugOutput::getErrorStream () << ": pDevice = " << *pDevice << std::endl;
#endif

   if (!pDevice)
   {
      DebugOutput::getErrorStream () << "Error: No Device was created!" << std::endl;

      return 4;
   }

   // @HACK - cause the device code to be executed.
   Enumeration *pEnum = pDevice->listDeviceJobPropertyKeys ();
   delete pEnum;

   // Check for an error
   if (pDevice->hasError ())
   {
      DebugOutput::getErrorStream () << "Error: The device has an error." << std::endl;

      return 5;
   }

   rc = pfnShowOptions (pDevice);

   if (pDevice->hasError ())
   {
      DebugOutput::getErrorStream () << "Error: The device has an error." << std::endl;

      return 6;
   }

   delete pDevice;

   if (hmodDevice)
   {
#ifndef RETAIL
      int rc2 =
#endif

                g_module_close (hmodDevice);

#ifndef RETAIL
      if (DebugOutput::shouldOutputDeviceTester ()) DebugOutput::getErrorStream () << "OmniDeviceOptions::" << __FUNCTION__ << ": g_module_close returns " << rc2 << std::endl;
#endif
   }

   return rc;
}

int
showDeviceGS (Device *pDevice)
{
   std::cout << "For gs add the following: -sproperties=\"...\"" << std::endl;
   std::cout << "\twhere you replace ... with the following space separated key=value (* - default):" << std::endl;
   std::cout << std::endl;

   Enumeration *pEnum = pDevice->listJobPropertyKeys ();

   while (pEnum->hasMoreElements ())
   {
      char *pszKey = (char *)pEnum->nextElement ();

      if (!pszKey)
         break;

      std::string *pstringValue = pDevice->getJobPropertyType (pszKey);

      if (pstringValue)
      {
         Enumeration            *pEnumValues        = pDevice->listKeyValues (pszKey);
         std::string::size_type  idxSpace1          = std::string::npos;
         std::string::size_type  idxSpace2          = std::string::npos;
         std::string             stringDefaultValue;

         idxSpace1 = pstringValue->find (" ");

         if (std::string::npos != idxSpace1)
         {
            idxSpace2 = pstringValue->find (" ", idxSpace1 + 1);
         }
         if (std::string::npos != idxSpace2)
         {
            stringDefaultValue = pstringValue->substr (idxSpace1 + 1,
                                                       idxSpace2 - idxSpace1 - 1);
         }
         else
         {
            stringDefaultValue = pstringValue->substr (idxSpace1 + 1);
         }

         if (pEnumValues->hasMoreElements ())
         {
            std::cout << pszKey << "=" << std::endl;
         }

         while (pEnumValues->hasMoreElements ())
         {
            char *pszValue = (char *)pEnumValues->nextElement ();

            std::cout << "\t" << pszValue;

            if (0 == stringDefaultValue.compare (pszValue))
            {
               std::cout << " (*)";
            }

            std::cout << std::endl;
         }

         delete pEnumValues;
         delete pstringValue;
      }
   }

   delete pEnum;

   pEnum = pDevice->listDeviceJobPropertyKeys ();

   while (pEnum->hasMoreElements ())
   {
      char *pszKey = (char *)pEnum->nextElement ();

      if (!pszKey)
         break;

      std::cout << pszKey << "=" << std::endl;

      std::string *pstringValue = pDevice->getJobPropertyType (pszKey);

      if (pstringValue)
      {
         std::string::size_type idxSpace1          = std::string::npos;
         std::string::size_type idxSpace2          = std::string::npos;
         std::string::size_type idxSpace3          = std::string::npos;
         std::string            stringDefaultValue;

         idxSpace1 = pstringValue->find (" ");

         if (std::string::npos != idxSpace1)
         {
            idxSpace2 = pstringValue->find (" ", idxSpace1 + 1);
         }
         if (std::string::npos != idxSpace2)
         {
            idxSpace3 = pstringValue->find (" ", idxSpace2 + 1);
         }

         if (  std::string::npos != idxSpace2
            && (  ! (  pstringValue->find ("string ") == 0
                    || pstringValue->find ("boolean ") == 0
                    )
               )
            )
         {
            stringDefaultValue = pstringValue->substr (idxSpace1 + 1,
                                                       idxSpace2 - idxSpace1 - 1);
         }
         else
         {
            stringDefaultValue = pstringValue->substr (idxSpace1 + 1);
         }

         if (  pstringValue->find ("string ") == 0
            || pstringValue->find ("boolean ") == 0
            )
         {
            Enumeration *pEnumValues = pDevice->listKeyValues (pszKey);

            while (pEnumValues->hasMoreElements ())
            {
               char *pszValue = (char *)pEnumValues->nextElement ();

               std::cout << "\t" << pszValue;

               if (0 == stringDefaultValue.compare (pszValue))
               {
                  std::cout << " (*)";
               }

               std::cout << std::endl;
            }

            delete pEnumValues;
         }
         else if (pstringValue->find ("integer ") == 0)
         {
            std::string stringMin = "1";
            std::string stringMax = "n";

            if (std::string::npos != idxSpace2)
            {
               if (std::string::npos != idxSpace3)
               {
                  stringMin = pstringValue->substr (idxSpace2 + 1,
                                                    idxSpace3 - idxSpace2 - 1);
               }
               else
               {
                  stringMin = pstringValue->substr (idxSpace2 + 1);
               }
            }
            if (std::string::npos != idxSpace3)
            {
               stringMax = pstringValue->substr (idxSpace3 + 1);
            }

            std::cout << "\t"
                      << stringMin
                      << " ... "
                      << stringMax
                      << " ("
                      << stringDefaultValue
                      << ")"
                      << std::endl;
         }
         else if (pstringValue->find ("float ") == 0)
         {
            std::string stringMin = "1.0";
            std::string stringMax = "n";

            if (std::string::npos != idxSpace2)
            {
               if (std::string::npos != idxSpace3)
               {
                  stringMin = pstringValue->substr (idxSpace2 + 1,
                                                    idxSpace3 - idxSpace2 - 1);
               }
               else
               {
                  stringMin = pstringValue->substr (idxSpace2 + 1);
               }
            }
            if (std::string::npos != idxSpace3)
            {
               stringMax = pstringValue->substr (idxSpace3 + 1);
            }

            std::cout << "\t"
                      << stringMin
                      << " ... "
                      << stringMax
                      << " ("
                      << stringDefaultValue
                      << ")"
                      << std::endl;
         }

         delete pstringValue;
      }
   }

   delete pEnum;

   return 0;
}

int
showDeviceVerbose (Device *pDevice)
{
   std::cout << "Driver version is : "
             << pDevice->getVersion ()
             << std::endl
             << "Driver name is : "
             << pDevice->getDriverName ()
             << std::endl
             << "Device name is : "
             << pDevice->getDeviceName ()
             << std::endl
             << "Device short name is : "
             << pDevice->getShortName ()
             << std::endl
             << "Device library name is : "
             << pDevice->getLibraryName ()
             << std::endl;

   std::cout << "Device class is : ";
   switch (pDevice->getOmniClass ())
   {
   case OMNI_CLASS_UNKNOWN:  std::cout << "*** UNKNOWN ***" << std::endl; break;
   case OMNI_CLASS_COMPILED: std::cout << "compiled" << std::endl; break;
   case OMNI_CLASS_XML:      std::cout << "XML" << std::endl; break;
   case OMNI_CLASS_UPDF:     std::cout << "UPDF" << std::endl; break;
   }

   std::cout << "The PDL level is "
             << pDevice->getPDLLevel ()
             << std::endl
             << "The PDL sublevel is "
             << pDevice->getPDLSubLevel ()
             << std::endl
             << "The PDL major revision level is "
             << pDevice->getPDLMajorRevisionLevel ()
             << std::endl
             << "The PDL minor revision level is "
             << pDevice->getPDLMinorRevisionLevel ()
             << std::endl;

   DeviceOrientation *pDO   = pDevice->getCurrentOrientation ();
   Enumeration       *pEnum = pDO->getEnumeration ();

   std::cout << "Current orientation is:" << std::endl;
   std::cout << DeviceOrientation::IDToName (pDO->getID ()) << std::endl;

   std::cout << "Enumerating orientations:" << std::endl;
   while (pEnum->hasMoreElements ())
   {
      pDO = (DeviceOrientation *)pEnum->nextElement ();

      std::cout << DeviceOrientation::IDToName (pDO->getID ()) << std::endl;

      delete pDO;
   }
   std::cout << "Done!" << std::endl;

   delete pEnum;

   DeviceForm  *pDF = pDevice->getCurrentForm ();
   pEnum = pDF->getEnumeration ();

   std::cout << "Current form is:" << std::endl;
   std::cout << DeviceForm::IDToName (pDF->getID ()) << std::endl;
   HardCopyCap *pHCC = pDF->getHardCopyCap ();
   std::cout << "Form dimensions are: ("
             << pHCC->getCx ()
             << ","
             << pHCC->getCy ()
             << ") (L:"
             << pHCC->getLeftClip ()
             << ",T:"
             << pHCC->getTopClip ()
             << ",R:"
             << pHCC->getRightClip ()
             << ",B:"
             << pHCC->getBottomClip ()
             << ")"
             << std::endl;

   std::cout << "Enumerating forms:" << std::endl;
   while (pEnum->hasMoreElements ())
   {
      pDF = (DeviceForm *)pEnum->nextElement ();

      std::cout << DeviceForm::IDToName (pDF->getID ()) << std::endl;

      delete pDF;
   }
   std::cout << "Done!" << std::endl;

   delete pEnum;

   DeviceTray *pDT = pDevice->getCurrentTray ();
   pEnum = pDT->getEnumeration ();

   std::cout << "Current tray is:" << std::endl;
   std::cout << DeviceTray::IDToName (pDT->getID ()) << std::endl;

   std::cout << "Enumerating trays:" << std::endl;
   while (pEnum->hasMoreElements ())
   {
      pDT = (DeviceTray *)pEnum->nextElement ();

      std::cout << DeviceTray::IDToName (pDT->getID ()) << std::endl;

      delete pDT;
   }
   std::cout << "Done!" << std::endl;

   delete pEnum;

   DeviceMedia *pDM = pDevice->getCurrentMedia ();
   pEnum = pDM->getEnumeration ();

   std::cout << "Current media is:" << std::endl;
   std::cout << DeviceMedia::IDToName (pDM->getID ()) << std::endl;

   std::cout << "Enumerating medias:" << std::endl;
   while (pEnum->hasMoreElements ())
   {
      pDM = (DeviceMedia *)pEnum->nextElement ();

      std::cout << DeviceMedia::IDToName (pDM->getID ()) << std::endl;

      delete pDM;
   }
   std::cout << "Done!" << std::endl;

   delete pEnum;

   DeviceResolution *pDR = pDevice->getCurrentResolution ();
   pEnum = pDR->getEnumeration ();

   std::cout << "Current resolution is:" << std::endl;
   std::cout << DeviceResolution::IDToName (pDR->getID ()) << std::endl;

   std::cout << "Resolution dimensions are: ("
             << pDR->getXRes ()
             << ","
             << pDR->getYRes ()
             << ")"
             << std::endl;

   std::cout << "Enumerating resolutions:" << std::endl;
   while (pEnum->hasMoreElements ())
   {
      pDR = (DeviceResolution *)pEnum->nextElement ();

      std::cout << DeviceResolution::IDToName (pDR->getID ()) << std::endl;

      delete pDR;
   }
   std::cout << "Done!" << std::endl;

   delete pEnum;

   DevicePrintMode *pDPM = pDevice->getCurrentPrintMode ();
   pEnum = pDPM->getEnumeration ();

   std::cout << "Current print mode is:" << std::endl;
   std::cout << DevicePrintMode::IDToName (pDPM->getID ()) << std::endl;

   std::cout << "Print Mode dimensions are: (CT:"
             << pDPM->getColorTech ()
             << ",P:"
             << pDPM->getPhysicalCount ()
             << ",L:"
             << pDPM->getLogicalCount ()
             << ",#planes:"
             << pDPM->getNumPlanes ()
             << ")"
             << std::endl;

   std::cout << "Enumerating print modes:" << std::endl;
   while (pEnum->hasMoreElements ())
   {
      pDPM = (DevicePrintMode *)pEnum->nextElement ();

      std::cout << DevicePrintMode::IDToName (pDPM->getID ()) << std::endl;

      delete pDPM;
   }
   std::cout << "Done!" << std::endl;

   delete pEnum;

   DeviceGamma *pDG = pDevice->getCurrentGamma ();

   std::cout << "Current gamma is:" << std::endl;

   if (pDG)
   {
      std::cout << "{ "
                << pDG->getCGamma ()
                << " "
                << pDG->getMGamma ()
                << " "
                << pDG->getYGamma ()
                << " "
                << pDG->getKGamma ()
                << " "
                << pDG->getCBias ()
                << " "
                << pDG->getMBias ()
                << " "
                << pDG->getYBias ()
                << " "
                << pDG->getKBias ()
                << "}"
                << std::endl;
   }
   else
   {
      std::cout << "Non existant!" << std::endl;
   }

   pEnum = DeviceDither::getEnumeration ();

///std::cout << "Current dither is:" << std::endl;
///std::cout << pDevice->getCurrentDither () << std::endl;

   std::cout << "Current dither id is:" << std::endl;
   std::cout << pDevice->getCurrentDitherID () << std::endl;

   std::cout << "Enumerating dithering ids:" << std::endl;
   while (pEnum->hasMoreElements ())
   {
      char *pszID = (char *)pEnum->nextElement ();

      std::cout << pszID << std::endl;
   }
   std::cout << "Done!" << std::endl;

   delete pEnum;

   pEnum = pDevice->listJobPropertyKeys ();

   std::cout << "Enumerating job property types:" << std::endl;
   while (pEnum->hasMoreElements ())
   {
      char *pszKey = (char *)pEnum->nextElement ();

      if (!pszKey)
      {
         DebugOutput::getErrorStream () << "Error: Enumeration returns null!" << std::endl;
         break;
      }

      std::string *pstringValue = pDevice->getJobPropertyType (pszKey);

      if (!pstringValue)
      {
         DebugOutput::getErrorStream () << "Error: getJobPropertyType (" << pszKey << ") return null!" << std::endl;
         break;
      }

      std::cout << pszKey << " = " << *pstringValue << std::endl;

      delete pstringValue;
   }
   std::cout << "Done!" << std::endl;

   delete pEnum;

   pEnum = pDevice->listDeviceJobPropertyKeys ();

   std::cout << "Enumerating device job property types:" << std::endl;
   while (pEnum->hasMoreElements ())
   {
      char *pszKey = (char *)pEnum->nextElement ();

      if (!pszKey)
      {
         DebugOutput::getErrorStream () << "Error: Enumeration returns null!" << std::endl;
         break;
      }

      std::string *pstringValue = pDevice->getJobPropertyType (pszKey);

      if (!pstringValue)
      {
         DebugOutput::getErrorStream () << "Error: getJobPropertyType (" << pszKey << ") return null!" << std::endl;
         break;
      }

      std::cout << pszKey << " = " << *pstringValue << std::endl;

      delete pstringValue;
   }
   std::cout << "Done!" << std::endl;

   delete pEnum;

   pEnum = pDevice->listJobPropertyKeys ();

   std::cout << "Enumerating job properties:" << std::endl;
   while (pEnum->hasMoreElements ())
   {
      char *pszKey = (char *)pEnum->nextElement ();

      if (!pszKey)
      {
         DebugOutput::getErrorStream () << "Error: Enumeration returns null!" << std::endl;
         break;
      }

      std::string *pstringValue = pDevice->getJobProperty (pszKey);

      if (!pstringValue)
      {
         DebugOutput::getErrorStream () << "Error: getJobProperty (" << pszKey << ") return null!" << std::endl;
         break;
      }

      std::cout << pszKey << " = " << *pstringValue << std::endl;

      delete pstringValue;
   }
   std::cout << "Done!" << std::endl;

   delete pEnum;

   pEnum = pDevice->listDeviceJobPropertyKeys ();

   std::cout << "Enumerating device job properties:" << std::endl;
   while (pEnum->hasMoreElements ())
   {
      char *pszKey = (char *)pEnum->nextElement ();

      if (!pszKey)
      {
         DebugOutput::getErrorStream () << "Error: Enumeration returns null!" << std::endl;
         break;
      }

      std::string *pstringValue = pDevice->getJobProperty (pszKey);

      if (!pstringValue)
      {
         DebugOutput::getErrorStream () << "Error: getJobProperty (" << pszKey << ") return null!" << std::endl;
         break;
      }

      std::cout << pszKey << " = " << *pstringValue << std::endl;

      delete pstringValue;
   }
   std::cout << "Done!" << std::endl;

   delete pEnum;

   PrintDevice *pPrintDevice = dynamic_cast <PrintDevice *>(pDevice);
   if (pPrintDevice)
   {
      std::string *pstringRet = 0;

      std::cout << "Job Properties in Device Values:" << std::endl;

      pstringRet = pPrintDevice->queryDeviceJobProperties ();

      if (pstringRet)
         std::cout << *pstringRet << std::endl;
      else
         std::cout << "NONE!" << std::endl;

      delete pstringRet;
   }

   if (pDevice->hasCapability (Capability::COLOR))
   {
      std::cout << "This device can print in color." << std::endl;
   }

   if (pDevice->hasCapability (Capability::MONOCHROME))
   {
      std::cout << "This device can print in monochrome." << std::endl;
   }

   if (pDevice->hasCapability (Capability::MIRROR))
   {
      std::cout << "This device can print in mirror image." << std::endl;
   }

   std::cout << "Portrait is ";
   pDO = pDevice->getCurrentOrientation ();
   if (!pDO->isOrientationSupported (DeviceOrientation::ORIENTATION_PORTRAIT))
   {
      std::cout << "not ";
   }
   std::cout << "supported" << std::endl;

   std::cout << "Letter is ";
   pDF = pDevice->getCurrentForm ();
   if (!pDF->isFormSupported (DeviceForm::FORM_LETTER))
   {
      std::cout << "not ";
   }
   std::cout << "supported" << std::endl;

   std::cout << "Auto is ";
   pDT = pDevice->getCurrentTray ();
   if (!pDT->isTraySupported (DeviceTray::TRAY_AUTO))
   {
      std::cout << "not ";
   }
   std::cout << "supported" << std::endl;

   std::cout << "Plain is ";
   pDM = pDevice->getCurrentMedia ();
   if (!pDM->isMediaSupported (DeviceMedia::MEDIA_PLAIN))
   {
      std::cout << "not ";
   }
   std::cout << "supported" << std::endl;

   std::cout << "360x360 is ";
   pDR = pDevice->getCurrentResolution ();
   if (!pDR->isResolutionSupported (DeviceResolution::RESOLUTION_360_X_360))
   {
      std::cout << "not ";
   }
   std::cout << "supported" << std::endl;

   std::cout << "CMYK is ";
   pDPM = pDevice->getCurrentPrintMode ();
   if (!pDPM->isPrintModeSupported (DevicePrintMode::PRINT_MODE_24_CMYK))
   {
      std::cout << "not ";
   }
   std::cout << "supported" << std::endl;

   return 0;
}

int
main (int argc, char *argv[])
{
   int            rc                = 0;
   bool           fUsePDC           = false;
   PFNSHOWOPTIONS pfnShowOptions    = showDeviceGS;
   PSZRO          pszJobProperties  = 0;
   JobProperties  jobProp;

   if (!g_module_supported ())
   {
      DebugOutput::getErrorStream () << "Error: This program needs glib's module routines!" << std::endl;

      return 1;
   }

   if (2 > argc)
   {
      printUsage (argv[0]);

      return 2;
   }

   Omni::initialize ();

   for (int i = 1; i < argc; i++)
   {
      if (0 == strcasecmp (argv[i], "--all"))
      {
         Enumeration *pEnum = Omni::listDevices ();

         while (pEnum->hasMoreElements ())
         {
            OmniDevice *pOD = (OmniDevice *)pEnum->nextElement ();

            if (pOD)
            {
               PSZCRO        pszLibName  = pOD->getLibraryName ();
               PSZRO         pszJobProps = pOD->getJobProperties ();
               JobProperties jobPropNew  = jobProp;

               jobPropNew.setJobProperties (pszJobProps);

               pszJobProps = jobPropNew.getJobProperties ();

               std::cerr << "Checking: " << pszLibName << ", \"" << pszJobProps << "\"" << std::endl;

               rc |= executeDevice (pszLibName,
                                    fUsePDC,
                                    pfnShowOptions,
                                    pszJobProps);

               if (pszJobProps)
               {
                  free ((void*)pszJobProps);
                  pszJobProps = 0;
               }

               delete pOD;
            }
            else
            {
               std::cerr << "Error" << std::endl;

               rc = 1;
            }
         }

         delete pEnum;
      }
      else if (0 == strcasecmp (argv[i], "--UsePDC"))
      {
         fUsePDC = true;
      }
      else if (0 == strcasecmp (argv[i], "--!UsePDC"))
      {
         fUsePDC = false;
      }
      else if (0 == strcasecmp (argv[i], "--Verbose"))
      {
         pfnShowOptions = showDeviceVerbose;
      }
      else if (0 == strcasecmp (argv[i], "--driver"))
      {
         std::ostringstream  oss;
         std::string         stringOss;
         char               *pszDriverLibrary = argv[i + 1];

         if (0 != strncasecmp (argv[i + 1], "lib", 3))
         {
            oss << "lib"
                << pszDriverLibrary
                << ".so";

            stringOss        = oss.str ();
            pszDriverLibrary = (char *)stringOss.c_str ();
         }

         rc |= executeDevice (pszDriverLibrary,
                              fUsePDC,
                              pfnShowOptions,
                              pszJobProperties);

         i++;
      }
      else if (0 == strncasecmp ("-sproperties=", argv[i], 13))
      {
         char *pszJobProp = argv[i] + 13;

         jobProp.setJobProperties (pszJobProp);

#ifndef RETAIL
         jobProp.applyAllDebugOutput ();
#endif

         if (pszJobProperties)
         {
            free ((void *)pszJobProperties);
            pszJobProperties = 0;
         }
         pszJobProperties = jobProp.getJobProperties ();
      }
      else
      {
         DeviceInfo *pDI = Omni::findDeviceInfoFromShortName (argv[i]);

         if (pDI)
         {
            Device     *pDevice    = 0;
            GModule    *hmodDevice = 0;
            OmniDevice *pOD        = 0;
            PSZRO       pszJobProp = 0;

            pDevice    = pDI->getDevice ();
            hmodDevice = pDI->getHmodDevice ();
            pOD        = pDI->getOmniDevice ();

            if (pOD)
            {
               pszJobProp = pOD->getJobProperties ();

               jobProp.setJobProperties (pszJobProp);

               pszJobProp = jobProp.getJobProperties ();
            }

            std::cout << "Info: For faster results use:"
                      << std::endl
                      << "Info:\t"
                      << argv[0];

            if (pszJobProp)
            {
               std::cout << " '-sproperties="
                         << pszJobProp
                         << "'";
            }

            std::cout << " --driver "
                      << pDevice->getLibraryName ()
                      << std::endl;

            rc |= executeDevice (pDevice->getLibraryName (),
                                 fUsePDC,
                                 pfnShowOptions,
                                 pszJobProp);

            if (pszJobProp)
            {
               free ((void *)pszJobProp);
               pszJobProp = 0;
            }

            delete pDI;
         }
      }
   }

   // Clean up!
   if (pszJobProperties)
   {
      free ((void *)pszJobProperties);
      pszJobProperties = 0;
   }

   Omni::terminate ();

   return rc;
}
