/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <XMLDeviceOrientation.hpp>

XMLDeviceOrientation::
XMLDeviceOrientation (Device      *pDevice,
                      int          id,
                      bool         fSimulationRequired,
                      xmlNodePtr   node)
   : DeviceOrientation (pDevice,
                        id,
                        fSimulationRequired)
{
   node_d        = node;
   pszDeviceID_d = 0;
}

XMLDeviceOrientation::
~XMLDeviceOrientation ()
{
   if (pszDeviceID_d)
   {
      free ((void *)pszDeviceID_d);
      pszDeviceID_d = 0;
   }
}

DeviceOrientation * XMLDeviceOrientation::
createV (int id)
{
   return XMLDeviceOrientation::create (pDevice_d, id);
}

DeviceOrientation * XMLDeviceOrientation::
create (Device *pDevice,
        PSZCRO  pszId)
{
   int id = DeviceOrientation::nameToID (pszId);

#ifndef RETAIL
   if (DebugOutput::shouldOutputXMLDeviceOrientation ()) DebugOutput::getErrorStream () << "XMLDeviceOrientation::" << __FUNCTION__ << ": id = " << id << std::endl;
#endif

   if (DeviceOrientation::ORIENTATION_UNLISTED != id)
   {
      return create (pDevice, id);
   }

   return 0;
}

DeviceOrientation * XMLDeviceOrientation::
create (Device *pDevice, int id)
{
   XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice);

   if (!pXMLDevice)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDeviceOrientation ()) DebugOutput::getErrorStream () << "XMLDeviceOrientation::" << __FUNCTION__ << ": !pXMLDevice " << std::endl;
#endif

      return 0;
   }

   xmlDocPtr          docDeviceOrientations  = pXMLDevice->getDocOrientations ();
   xmlNodePtr         rootDeviceOrientations = xmlDocGetRootElement (docDeviceOrientations);
   xmlNodePtr         elmDeviceOrientation   = 0;
   DeviceOrientation *pDORet                 = 0;

   if (!rootDeviceOrientations)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDeviceOrientation ()) DebugOutput::getErrorStream () << "XMLDeviceOrientation::" << __FUNCTION__ << ": !rootDeviceOrientations " << std::endl;
#endif

      return 0;
   }

   elmDeviceOrientation = firstNode (rootDeviceOrientations);
   if (!elmDeviceOrientation)
   {
#ifndef RETAIL
      if (DebugOutput::shouldOutputXMLDeviceOrientation ()) DebugOutput::getErrorStream () << "XMLDeviceOrientation::" << __FUNCTION__ << ": !elmDeviceOrientation " << std::endl;
#endif

      return 0;
   }

   elmDeviceOrientation = firstNode (elmDeviceOrientation->children);

   while (  elmDeviceOrientation
         && !pDORet
         )
   {
      PSZCRO pszOrientation = getXMLContentString (elmDeviceOrientation,
                                                   docDeviceOrientations,
                                                   "name");

      if (pszOrientation)
      {
         int idOrientation = DeviceOrientation::nameToID (pszOrientation);

         if (idOrientation == id)
         {
            bool fSimulationRequired = false; // @TBD

#ifndef RETAIL
            if (DebugOutput::shouldOutputXMLDeviceOrientation ()) DebugOutput::getErrorStream () << "XMLDeviceOrientation::" << __FUNCTION__ << ": Creating \"" << pszOrientation << "\"" << std::endl;
#endif

            // Create the object
            pDORet = new XMLDeviceOrientation (pDevice,
                                               id,
                                               fSimulationRequired,
                                               elmDeviceOrientation);
         }

         free ((void *)pszOrientation);
      }

      elmDeviceOrientation = nextNode (elmDeviceOrientation);
   }


#ifdef DEBUG
   if (DebugOutput::shouldOutputXMLDeviceOrientation ()) DebugOutput::getErrorStream () << "XMLDeviceOrientation::" << __FUNCTION__ << ": returning " << pDORet << std::endl;
#endif

   return pDORet;
}

PSZCRO XMLDeviceOrientation::
getDeviceID ()
{
   if (!pszDeviceID_d)
   {
      if (node_d)
      {
         pszDeviceID_d = getXMLContentString (node_d, node_d->doc, "deviceID");
      }
   }

#ifdef DEBUG
   if (DebugOutput::shouldOutputXMLDeviceForm ()) DebugOutput::getErrorStream () << "XMLDeviceOrientation::" << __FUNCTION__ << ": returning " << (pszDeviceID_d ? pszDeviceID_d : "NULL") << std::endl;
#endif

   return pszDeviceID_d;
}

bool XMLDeviceOrientation::
isOrientationSupported (int id)
{
   XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice_d);

   if (!pXMLDevice)
      return false;

   xmlDocPtr   docDeviceOrientations  = pXMLDevice->getDocOrientations ();
   xmlNodePtr  rootDeviceOrientations = xmlDocGetRootElement (docDeviceOrientations);
   xmlNodePtr  elmDeviceOrientation   = 0;
   bool        fFound                 = false;

   if (!rootDeviceOrientations)
      return false;

   elmDeviceOrientation = firstNode (rootDeviceOrientations);
   if (!elmDeviceOrientation)
      return false;

   elmDeviceOrientation = firstNode (elmDeviceOrientation->children);

   while (  elmDeviceOrientation
         && !fFound
         )
   {
      PSZCRO pszOrientation = getXMLContentString (elmDeviceOrientation,
                                                   docDeviceOrientations,
                                                   "name");

      if (pszOrientation)
      {
         int idOrientation = DeviceOrientation::nameToID (pszOrientation);

         if (idOrientation == id)
         {
            fFound = true;
         }

         free ((void *)pszOrientation);
      }

      elmDeviceOrientation = nextNode (elmDeviceOrientation);
   }

   return fFound;
}

class XMLOrientation_Enumerator : public Enumeration
{
public:
   XMLOrientation_Enumerator (Device *pDevice, xmlNodePtr nodeItem)
   {
      XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice);

      pXMLDevice_d            = pXMLDevice;
      docDeviceOrientations_d = 0;
      nodeItem_d              = nodeItem;

      if (pXMLDevice)
      {
         docDeviceOrientations_d = pXMLDevice->getDocOrientations ();
      }
      else
      {
         nodeItem_d = 0;
      }
   }

   virtual bool hasMoreElements ()
   {
      return nodeItem_d ? true : false;
   }

   virtual void *nextElement ()
   {
      DeviceOrientation *pDORet = 0;

      if (nodeItem_d)
      {
         PSZCRO pszOrientationName = getXMLContentString (nodeItem_d,
                                                          docDeviceOrientations_d,
                                                          "name");

         if (pszOrientationName)
         {
            pDORet = XMLDeviceOrientation::create (pXMLDevice_d, pszOrientationName);

            free ((void *)pszOrientationName);
         }

         nodeItem_d = nextNode (nodeItem_d);
      }

      return pDORet;
   }

private:
   XMLDevice  *pXMLDevice_d;
   xmlDocPtr   docDeviceOrientations_d;
   xmlNodePtr  nodeItem_d;
};

Enumeration * XMLDeviceOrientation::
getEnumeration ()
{
   XMLDevice *pXMLDevice = XMLDevice::isAXMLDevice (pDevice_d);

   if (!pXMLDevice)
      return new XMLOrientation_Enumerator (pDevice_d, 0);

   xmlDocPtr   docDeviceOrientations  = pXMLDevice->getDocOrientations ();
   xmlNodePtr  rootDeviceOrientations = xmlDocGetRootElement (docDeviceOrientations);
   xmlNodePtr  elmDeviceOrientation   = 0;

   if (!rootDeviceOrientations)
      return new XMLOrientation_Enumerator (pDevice_d, 0);

   elmDeviceOrientation = firstNode (rootDeviceOrientations);
   if (!elmDeviceOrientation)
      return new XMLOrientation_Enumerator (pDevice_d, 0);

   elmDeviceOrientation = firstNode (elmDeviceOrientation->children);

   return new XMLOrientation_Enumerator (pDevice_d, elmDeviceOrientation);
}

std::string XMLDeviceOrientation::
toString (std::ostringstream& oss)
{
   std::ostringstream oss2;

   oss << "{"
       << DeviceOrientation::toString (oss2)
       << "}";

   return oss.str ();
}

/* Provide a way to print out class data
*/
std::ostream&
operator<< (std::ostream& os, const XMLDeviceOrientation& const_self)
{
   XMLDeviceOrientation& self = const_cast<XMLDeviceOrientation&>(const_self);
   std::ostringstream    oss;

   os << self.toString (oss);

   return os;
}
